import { Cookie } from "./Cookie";

if (typeof browser == "undefined") {
  var browser = chrome;
}

/**
 * Handles changes in the search settings when the user
 * makes them on our Website. It will catch the regular responses 
 * which would store the settings in form of cookies and stores
 * them within this web-extension instead.
 * Also it will attach settings that are stored in this extension
 * and attaches them to requests made to our website
 */
export class SettingsManager {
  initialized = false;

  store_settings = true;
  settings = {};

  constructor() { }

  async init() {
    if (this.initialized) return;
    return browser.storage.sync
      .get({ settings_store: true, settings_settings: {} })
      .then((stored_settings) => {
        this.store_settings = stored_settings.settings_store;
        this.settings = stored_settings.settings_settings;
        return this.updateSettingRule();
      }).then(() => {
        this.initialized = true;
        return this;
      });
  }

  /**
   * Merges new settings into existing ones and submits them
   * into storage
   * 
   * @param {object} settings 
   */
  async set(settings) {
    await this.init();
    this.settings = { ...this.settings, ...settings };
    return this.sync();
  }

  /**
   * Removes sepecified setting key 
   * and syncs settings
   * 
   * @param {string} setting_key 
   */
  async remove(setting_key) {
    await this.init();
    delete this.settings[setting_key];
    return this.sync();
  }

  /**
   * Syncs current setting state with storage
   */
  async sync() {
    return browser.storage.sync.set({ settings_store: this.store_settings, settings_settings: this.settings }).then(() => this.updateSettingRule());
  }

  async handleStorageChange(changes) {
    if (
      changes.hasOwnProperty("settings_store") &&
      typeof changes.settings_store.newValue == "boolean"
    ) {
      let new_value = changes.settings_store.newValue;
      if (new_value == true) {
        // Setting storage now enabled
        await Cookie.COOKIES_TO_SETTINGS();
      } else if (new_value == false) {
        // Setting storage now disabled. Put all settings into browser cookie storage
        await Cookie.SETTINGS_TO_COOKIES();
      }
      this.store_settings = changes.settings_store.newValue;
    }
    if (changes.hasOwnProperty("settings_settings")) {
      this.settings = changes.settings_settings.newValue;
      this.updateSettingRule();
    }
  }

  /**
   *
   * @param {*} details
   */
  async handleResponseHeaders(details) {
    if (!this.store_settings) return details.responseHeaders;
    await this.init();
    if (details.hasOwnProperty("responseHeaders")) {
      let settings_changed = false;
      let cookies_to_remove = [];
      for (const header of details.responseHeaders) {
        if (header.name.match(/set-cookie/i)) {
          let cookie_responses = header.value.split("\n");
          let new_cookies = [];
          for (let cookie of cookie_responses) {
            let response_cookie = new Cookie(cookie);

            if (response_cookie.expired) {
              delete this.settings[response_cookie.key];
              // Remove cookie from browser
              cookies_to_remove.push(response_cookie);
            } else if (["key", "cost", "tokens"].includes(response_cookie.key)) {
              // Do not handle those settings as tokenmanager will do so
              continue;
            } else {
              this.settings[response_cookie.key] = decodeURIComponent(
                response_cookie.value
              );
              cookies_to_remove.push(response_cookie);
            }
          }
          settings_changed = true;
        }
      }
      if (settings_changed) {
        await this.sync().then(() => {
          let removal_promises = [];
          for (let cookie of cookies_to_remove) {
            removal_promises.push(Cookie.REMOVE_COOKIE(cookie));
          }
          return Promise.all(removal_promises);
        });
      }
      return;
    }
  }

  /**
   * Creates a declarativeNetRequest Rule
   * for attaching the settings to requests to our website
   */
  async updateSettingRule() {
    if (!browser.hasOwnProperty("declarativeNetRequest")) return;
    let cookies = [];
    let ignore_settings = ["key", "cost"];
    for (let key in this.settings) {
      if (ignore_settings.includes(key)) continue;
      cookies.push(key + "=" + encodeURIComponent(this.settings[key]));
    }
    let rules = [];
    if (cookies.length > 0) {
      let cookie_string = cookies.join("; ");
      rules.push({
        id: 1,
        condition: {
          requestDomains: [
            "metager.de",
            "metager.org"
          ],
          resourceTypes: ["main_frame", "sub_frame", "xmlhttprequest"]
        },
        action: {
          type: "modifyHeaders",
          requestHeaders: [{
            header: "Cookie",
            operation: "append",
            value: cookie_string
          }]
        }
      });
    }
    return browser.declarativeNetRequest.updateDynamicRules({
      addRules: rules,
      removeRuleIds: [1]
    });
  }
}
